package org.xbib.interlibrary.api.action;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.SynchronousQueue;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import org.xbib.interlibrary.api.action.worker.WorkerThreadFactory;
import org.xbib.interlibrary.api.action.worker.WorkerThreadPoolExecutor;
import org.xbib.settings.Settings;

public abstract class AbstractAction<Req extends Request, Resp extends Response<Req>, B extends ResponseBuilder<Req, Resp>, Serv extends Service<Req, Resp, B>, Prov extends ServiceProvider<Serv>>
        implements Action<Req, Resp, B, Serv> {

    private static final Logger logger = Logger.getLogger(AbstractAction.class.getName());

    public AbstractAction() {
    }

    @Override
    public List<Resp> all(Req request) {
        try {
            List<CompletableFuture<B>> futures = submit(request);
            if (futures.isEmpty()) {
                return List.of(newEmptyResponseBuilder(request).build());
            }
            return all(futures).get().stream().map(ResponseBuilder::build).collect(Collectors.toList());
        } catch (Throwable t) {
            logger.log(Level.SEVERE, t.getMessage(), t);
            return List.of(newErrorResponse(request, t).build());
        }
    }

    @Override
    public Resp any(Req request) {
        try {
            List<CompletableFuture<B>> futures = submit(request);
            if (futures.isEmpty()) {
                return newEmptyResponseBuilder(request).build();
            }
            Optional<Resp> optionalResp = any(futures).get().stream().map(ResponseBuilder::build).findFirst();
            return optionalResp.orElse(newEmptyResponseBuilder(request).build());
        } catch (Throwable t) {
            logger.log(Level.SEVERE, t.getMessage(), t);
            return newErrorResponse(request, t).build();
        }
    }

    @Override
    public Resp reduce(Req request) {
        try {
            List<CompletableFuture<B>> futures = submit(request);
            if (futures.isEmpty()) {
                return newEmptyResponseBuilder(request).build();
            }
            return all(futures).get().stream().reduce((first, second) -> second).map(ResponseBuilder::build).orElse(null);
        } catch (Throwable t) {
            logger.log(Level.SEVERE, t.getMessage(), t);
            return newErrorResponse(request, t).build();
        }
    }

    @Override
    public Resp allWithRetry(Req request) {
        try {
            List<CompletableFuture<B>> futures = submit(request);
            if (futures.isEmpty()) {
                return newEmptyResponseBuilder(request).build();
            }
            Map<Response.Status, List<B>> map = all(futures).get().stream()
                    .collect(Collectors.groupingBy(ResponseBuilder::getStatus));
            // sort groups
            List<B> list = new ArrayList<>();
            for (Response.Status status : List.of(Response.Status.OK,
                    Response.Status.RETRY,
                    Response.Status.EMPTY_RESPONSE,
                    Response.Status.NOT_FOUND,
                    Response.Status.TIMEOUT,
                    Response.Status.ERROR,
                    Response.Status.UNDEFINED)) {
                if (map.containsKey(status)) {
                    list.addAll(map.get(status));
                }
            }
            Optional<B> optional = list.stream().findFirst();
            return optional.map(ResponseBuilder::build).orElse(null);
        } catch (Throwable t) {
            logger.log(Level.SEVERE, t.getMessage(), t);
            return newErrorResponse(request, t).build();
        }
    }

    @Override
    public final List<CompletableFuture<B>> submit(Req request) {
        List<CompletableFuture<B>> futures = new ArrayList<>();
        List<Serv> services = getServices().stream()
                .filter(service -> isServiceRequested(request, service))
                .filter(service -> isServiceQualified(request, service))
                .filter(Service::isEnabled)
                .filter(this::isServiceHealthy)
                .toList();
        if (services.isEmpty()) {
            newErrorResponse(request, new IllegalArgumentException("no services available for execution"));
        } else {
           WorkerThreadPoolExecutor executor = new WorkerThreadPoolExecutor(services.size(),
                    new SynchronousQueue<>(),
                    new WorkerThreadFactory("interlibrary-action"));
            Req validatedRequest = validate(request);
            B responseBuilder = newResponseBuilder(validatedRequest);
            for (Serv service : services) {
                futures.add(CompletableFuture.supplyAsync(() -> {
                    executeService(service, validatedRequest, responseBuilder);
                    return responseBuilder;
                }, executor));
            }
            executor.shutdown();
        }
        return futures;
    }

    @Override
    public void close() throws IOException {
        if (getServices() == null) {
            return;
        }
        for (Service<?, ?, ?> service : getServices()) {
            try {
                service.close();
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        }
    }

    @Override
    public Req validate(Req request) {
        return request;
    }

    protected boolean isServiceRequested(Req request, Serv service) {
        return request.getServices() == null ||
                service.getName() == null ||
                request.getServices().isEmpty() ||
                request.getServices().contains(service.getName());
    }

    protected boolean isServiceQualified(Req request, Serv service) {
        return true;
    }

    protected boolean isServiceHealthy(Serv service) {
        return true;
    }

    @SuppressWarnings("unchecked")
    protected List<Serv> createServices(String serviceGroup, ActionArguments arguments) {
        Settings settings = arguments.getSettings();
        ClassLoader classLoader = arguments.getClassLoader();
        List<Serv> services = new ArrayList<>();
        for (Map.Entry<String, Settings> serviceEntry : settings.getGroups(serviceGroup).entrySet()) {
            String serviceName = serviceEntry.getKey();
            Settings serviceSettings = serviceEntry.getValue();
            if (serviceSettings.getAsBoolean("enabled", true)) {
                try {
                    Class<Prov> clazz = (Class<Prov>) classLoader.loadClass(serviceSettings.get("class"));
                    Prov provider = clazz.getDeclaredConstructor().newInstance();
                    ServiceArguments serviceArguments = new DefaultServiceArguments(settings,
                             serviceName, serviceSettings, arguments.getInterlibraryConfiguration());
                    Serv service = provider.createService(serviceArguments);
                    services.add(service);
                } catch (Exception e) {
                    if (settings.getAsBoolean("strict", false)) {
                        throw new IllegalArgumentException(e);
                    } else {
                        logger.log(Level.WARNING, "service not initializable: " + serviceName + " " + e.getMessage(), e);
                    }
                }
            } else {
                logger.log(Level.WARNING, "service disabled in configuration: " + serviceName);
            }
        }
        return services;
    }

    private void executeService(Serv service, Req request, B responseBuilder) {
        try {
            service.execute(request, responseBuilder);
        } catch (Throwable t) {
            responseBuilder.setThrowable(service.getName(), t.getMessage(), t);
            service.addThrowable(t);
        }
    }

    private static <T> CompletableFuture<List<T>> all(List<CompletableFuture<T>> futures) {
        CompletableFuture<?>[] cfs = futures.toArray(new CompletableFuture<?>[0]);
        return CompletableFuture.allOf(cfs).thenApply(ignored -> futures.stream()
                .map(CompletableFuture::join)
                .collect(Collectors.toList()));
    }

    private static <T> CompletableFuture<List<T>> any(List<CompletableFuture<T>> futures) {
        CompletableFuture<?>[] cfs = futures.toArray(new CompletableFuture<?>[0]);
        return CompletableFuture.anyOf(cfs).thenApply(ignored -> futures.stream()
                .map(CompletableFuture::join)
                .collect(Collectors.toList()));
    }
}
