package org.xbib.interlibrary.api.action.worker;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * A worker pool for processing request by a number of worker threads.
 * If worker threads exit early, they are removed and finished, not reused.
 * If no worker is left, the pool closes.
 */
public abstract class AbstractWorkerPool<R> implements WorkerPool<R>, AutoCloseable {

    private static final Logger logger = Logger.getLogger(AbstractWorkerPool.class.getName());

    private static final int DEFAULT_WAIT_SECONDS = 30;

    private final BlockingQueue<WorkerRequest<R>> queue;

    private final ThreadPoolWorkerExecutor executor;

    private final int workerCount;

    private final int waitSeconds;

    protected final AtomicBoolean closed;

    private final CountDownLatch active;

    private final CountDownLatch passive;

    private final Map<Runnable, Throwable> exceptions;

    private final WorkerPoolListener<WorkerPool<R>> listener;

    private final AtomicLong executionCounter;

    private volatile int returncode;

    public AbstractWorkerPool(String name,
                              int workerCount,
                              WorkerPoolListener<WorkerPool<R>> listener) {
        this(name, workerCount, listener, DEFAULT_WAIT_SECONDS);
    }

    public AbstractWorkerPool(String name,
                              int workerCount,
                              WorkerPoolListener<WorkerPool<R>> listener,
                              int waitSeconds) {
        this.workerCount = workerCount;
        this.waitSeconds = waitSeconds;
        this.listener = listener;
        this.queue = new SynchronousQueue<>(true);
        this.closed = new AtomicBoolean(true);
        this.active = new CountDownLatch(workerCount);
        this.passive = new CountDownLatch(workerCount);
        this.exceptions = new ConcurrentHashMap<>();
        this.executionCounter = new AtomicLong();
        this.executor = new ThreadPoolWorkerExecutor(workerCount,
                new LinkedBlockingQueue<>(),
                new WorkerThreadFactory(name));
        this.returncode = 0;
        logger.log(Level.INFO, () -> MessageFormat.format("pool: worker count = {0}, wait seconds = {1}",
                workerCount, waitSeconds));
    }

    @Override
    public void open() throws IOException {
        if (closed.compareAndSet(true, false)) {
            for (int i = 0; i < workerCount; i++) {
                String name = "worker-" + i + "-" + workerCount;
                Worker<R> worker = newWorker(name, i);
                WrappedWorker wrappedWorker = new WrappedWorker(this, worker);
                executor.submit(wrappedWorker);
            }
            // we are after openInput(), push all inputs to workers
            for (WorkerRequest<R> workerRequest : this) {
                submit(workerRequest);
            }
        }
    }

    @Override
    public AtomicLong getExecutionCounter() {
        return executionCounter;
    }

    @Override
    public BlockingQueue<WorkerRequest<R>> getQueue() {
        return queue;
    }

    public Map<Runnable, Throwable> getExceptions() {
        return exceptions;
    }

    @Override
    public void setReturnCode(int returncode) {
        this.returncode = returncode;
    }

    @Override
    public int getReturnCode() {
        return returncode;
    }

    @Override
    public void submit(WorkerRequest<R> request) {
        if (request == null) {
            // ignore
            return;
        }
        if (request.isPoison()) {
            // ignore silently
            return;
        }
        if (closed.get()) {
            returncode = 1;
            // already closed
            throw new WorkerPoolException("closed, request " + request + " rejected");
        }
        if (active.getCount() == 0 && passive.getCount() == 0) {
            returncode = 1;
            logger.log(Level.SEVERE, () -> "no more workers available, closing pool");
            try {
                close();
            } catch (IOException e) {
                // ignore
            }
            WorkerPoolException e = new WorkerPoolException("no more workers available");
            exceptions.put(Thread.currentThread(), e);
            throw e;
        }
        try {
            queue.put(request);
            executionCounter.incrementAndGet();
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            returncode = 1;
            try {
                close();
            } catch (IOException ioe) {
                // ignore
            }
            throw new WorkerPoolException(e.getMessage());
        }
    }

    @Override
    public void close() throws IOException {
        if (closed.compareAndSet(false, true)) {
            if (active.getCount() == 0) {
                while (passive.getCount() > 0) {
                    logger.log(Level.INFO, () -> MessageFormat.format("closing pool, sending poison, {0} workers", passive.getCount()));
                    try {
                        queue.put(new WorkerRequest<>() {
                            @Override
                            public R getRequest() {
                                return null;
                            }

                            @Override
                            public boolean isPoison() {
                                return true;
                            }

                            @Override
                            public String getRequestOrigin() {
                                return null;
                            }
                        });
                        Thread.sleep(50L); // important, wait to let receiver thread finish and adjust latch
                    } catch (InterruptedException e) {
                        Thread.currentThread().interrupt();
                    }
                }
            } else {
                logger.log(Level.SEVERE, () -> "closing, not all workers were active, count = " + (workerCount - active.getCount()));
            }
            logger.log(Level.INFO, () -> "now shutting down executor, return code " + returncode);
            try {
                executor.shutdown();
                executor.awaitTermination(waitSeconds, TimeUnit.SECONDS);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                throw new IOException(e);
            } finally {
                if (listener != null) {
                    if (exceptions.isEmpty()) {
                        listener.success(this);
                    } else {
                        listener.failure(this, exceptions);
                    }
                }
            }
        }
    }

    private class ThreadPoolWorkerExecutor extends ThreadPoolExecutor {

        private ThreadPoolWorkerExecutor(int numThreads,
                                         BlockingQueue<Runnable> blockingQueue,
                                         ThreadFactory threadFactory) {
            super(numThreads, numThreads, 0L, TimeUnit.MILLISECONDS,
                    blockingQueue, threadFactory);
        }

        @Override
        protected void afterExecute(Runnable runnable, Throwable terminationCause) {
            super.afterExecute(runnable, terminationCause);
            Throwable throwable = terminationCause;
            if (throwable == null && runnable instanceof Future<?>) {
                try {
                    Future<?> future = (Future<?>) runnable;
                    if (future.isDone()) {
                        future.get();
                    }
                } catch (CancellationException ce) {
                    logger.log(Level.FINEST, ce.getMessage(), ce);
                    throwable = ce;
                } catch (ExecutionException ee) {
                    logger.log(Level.FINEST, ee.getMessage(), ee);
                    throwable = ee.getCause();
                } catch (InterruptedException ie) {
                    Thread.currentThread().interrupt();
                    logger.log(Level.FINEST, ie.getMessage(), ie);
                }
            }
            if (throwable != null) {
                logger.log(Level.SEVERE, throwable.getMessage(), throwable);
                exceptions.put(runnable, throwable);
                setReturnCode(1);
            }
        }
    }

    private static class WorkerThreadFactory implements ThreadFactory {

        private final String name;

        private final AtomicInteger counter;

        WorkerThreadFactory(String name) {
            this.name = name;
            this.counter = new AtomicInteger();
        }

        @Override
        public Thread newThread(Runnable r) {
            Thread thread = new Thread(r, name + "-" + counter.getAndIncrement());
            thread.setDaemon(true);
            return thread;
        }
    }

    private class WrappedWorker implements Runnable {

        private final AbstractWorkerPool<R> pool;

        private final Worker<R> worker;

        private final Logger logger;

        private WrappedWorker(AbstractWorkerPool<R> pool, Worker<R> worker) {
            this.pool = pool;
            this.worker = worker;
            this.logger = Logger.getLogger(worker.getClass().getName());
        }

        @Override
        public void run() {
            WorkerRequest<R> request = null;
            try {
                logger.log(Level.INFO, () -> MessageFormat.format("start of worker {0}", worker));
                // worker is now active
                active.countDown();
                while (true) {
                    request = getQueue().take();
                    if (request.isPoison()) {
                        break;
                    }
                    try {
                        worker.execute(request);
                    } finally {
                        pool.afterExecute(worker, request);
                    }
                }
            } catch (InterruptedException e) {
                // we got interrupted, this may lead to data loss. Clear interrupt state and log warning.
                Thread.currentThread().interrupt();
                logger.log(Level.WARNING, e.getMessage(), e);
                exceptions.put(this, e);
            } catch (Exception | AssertionError e) {
                // catch unexpected exceptions. Throwables, Errors are examined in afterExecute.
                logger.log(Level.SEVERE, e.getMessage(), e);
                // if we are late to the party, try to get the poison pill here
                if (closed.get()) {
                    try {
                        getQueue().poll(1, TimeUnit.MINUTES);
                    } catch (InterruptedException e2) {
                        Thread.currentThread().interrupt();
                        logger.log(Level.WARNING, e2.getMessage(), e2);
                    }
                }
                exceptions.put(this, e);
                throw new UncheckedIOException(e instanceof IOException ? (IOException) e : new IOException(e));
            } finally {
                // worker is now passive
                passive.countDown();
                if (exceptions.containsKey(this) && passive.getCount() == 0) {
                    logger.log(Level.INFO, "last worker exited with error, draining queue");
                    Collection<WorkerRequest<R>> collection = new ArrayList<>();
                    getQueue().drainTo(collection);
                }
                if (request != null && request.isPoison()) {
                    logger.log(Level.INFO, () -> MessageFormat.format("end of worker {0} {1}",
                            worker, "(completed, " + getExecutionCounter().get() + " requests)"));
                } else {
                    logger.log(Level.SEVERE, () -> MessageFormat.format("end of worker {0} {1}",
                            worker, "(abnormal termination after " + getExecutionCounter().get() + " requests)"));
                }
            }
        }
    }
}
