package org.xbib.interlibrary.api.library;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.datastructures.tiny.TinyMultiMap;
import org.xbib.interlibrary.api.config.Configuration;
import org.xbib.interlibrary.api.util.Maps;

public class DefaultLibrary implements Library, Comparable<Library> {

    private final LibraryArguments arguments;

    private final Collection<String> self;

    private final Collection<String> local;

    private final Collection<String> prefer;

    private final Collection<String> accept;

    private String isil;

    private String sigel;

    private String ezbId;

    private Collection<String> aliases;

    private String name;

    private String fullName;

    private Domain domain;

    private Place place;

    private Integer priority;

    private boolean isRegisteredInterlibrary;

    private boolean isCopyInterlibrary;

    private boolean isOnlineInterlibrary;

    private boolean isSubLibrary;

    private boolean isLastResort;

    private boolean isOrganization;

    private boolean isPublic;

    private Collection<String> groups;

    private Collection<Station> stations;

    private Predicate<Library> selfPredicate;

    private Predicate<Library> localPredicate;

    private Predicate<LibraryService> servicesPredicate;

    private Predicate<Library> groupPredicate;

    private LibrarySenderAddress senderAddress;

    private LibraryNotificationAddress notificationAddress;

    private LibraryPostalAddress postalAddress;

    private String activeEmail;

    private String passiveEmail;

    private String reservationEmail;

    protected DefaultLibrary() {
        this.arguments = null;
        this.self = new LinkedHashSet<>();
        this.local = new LinkedHashSet<>();
        this.accept = new LinkedHashSet<>();
        this.prefer = new LinkedHashSet<>();
        this.selfPredicate = library -> this.self.contains(library.getISIL());
        this.stations = new ArrayList<>();
    }

    public DefaultLibrary(LibraryArguments arguments) {
        this.arguments = Objects.requireNonNull(arguments);
        this.groupPredicate = library -> true;
        this.servicesPredicate = library -> true;
        this.self = new LinkedHashSet<>();
        this.local = new LinkedHashSet<>();
        this.accept = new LinkedHashSet<>();
        this.prefer = new LinkedHashSet<>();
        this.selfPredicate = library -> this.self.contains(library.getISIL());
        this.stations = new ArrayList<>();
        parseInternalMap(arguments.getMap());
        finishConfiguration(arguments.getInterlibraryConfiguration());
    }

    protected void finishConfiguration(Configuration configuration) {
        // recover ISIL
        String isil = arguments.getISIL();
        if (configuration != null) {
            configuration.getSelf(isil).forEach(s -> {
                if (s != null && !s.isEmpty()) {
                    self.add(s);
                }
            });
            configuration.getLocal(isil).forEach(s -> {
                if (s != null && !s.isEmpty()) {
                    local.add(s);
                }
            });
            configuration.getAccept(isil).forEach(s -> {
                if (s != null && !s.isEmpty()) {
                    accept.add(s);
                }
            });
            configuration.getPrefer(isil).forEach(s -> {
                if (s != null && !s.isEmpty()) {
                    prefer.add(s);
                }
            });
            // override from settings
            this.isRegisteredInterlibrary = configuration.isRegisteredInterlibrary(isil, isRegisteredInterlibrary());
            this.isCopyInterlibrary = configuration.isCopyInterlibrary(isil, isCopyInterlibrary());
            this.isOnlineInterlibrary = configuration.isOnlineInterlibrary(isil, isOnlineInterlibrary());
            this.isSubLibrary = configuration.isSubLibrary(isil, isSubLibrary());
            this.isLastResort = configuration.isLastResortLibrary(isil, isLastResort());
            this.isOrganization = configuration.isOrganization(isil, isOrganization());
            this.isPublic = configuration.isPublic(isil, isPublic());
        }
    }

    @Override
    public void setISIL(String isil) {
        this.isil = isil;
        if (isil != null && !isil.isEmpty()) {
            this.self.add(isil);
        }
    }

    @Override
    public String getISIL() {
        return isil;
    }

    @Override
    public void setSigel(String sigel) {
        this.sigel = sigel;
    }

    @Override
    public String getSigel() {
        return sigel;
    }

    @Override
    public void setEzbId(String ezbId) {
        this.ezbId = ezbId;
    }

    @Override
    public String getEzbId() {
        return ezbId;
    }

    @Override
    public void setAliases(Collection<String> aliases) {
        this.aliases = aliases;
    }

    @Override
    public Collection<String> getAliases() {
        return aliases;
    }

    @Override
    public void setName(String name) {
        this.name = name;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public void setFullName(String name) {
        this.fullName = name;
    }

    @Override
    public String getFullName() {
        return fullName;
    }

    @Override
    public void setRegisteredInterlibrary(boolean isRegisteredInterlibrary) {
        this.isRegisteredInterlibrary = isRegisteredInterlibrary;
    }

    @Override
    public boolean isRegisteredInterlibrary() {
        return isRegisteredInterlibrary;
    }

    @Override
    public void setCopyInterlibrary(boolean isCopyInterlibrary) {
        this.isCopyInterlibrary = isCopyInterlibrary;
    }

    @Override
    public boolean isCopyInterlibrary() {
        return isCopyInterlibrary;
    }

    @Override
    public void setOnlineInterlibrary(boolean isOnlineInterlibrary) {
        this.isOnlineInterlibrary = isOnlineInterlibrary;
    }

    @Override
    public boolean isOnlineInterlibrary() {
        return isOnlineInterlibrary;
    }

    @Override
    public void setLastResort(boolean isLastResort) {
        this.isLastResort = isLastResort;
    }

    @Override
    public boolean isLastResort() {
        return isLastResort;
    }

    public void setOrganization(boolean isOrganization) {
        this.isOrganization = isOrganization;
    }

    @Override
    public boolean isOrganization() {
        return isOrganization;
    }

    @Override
    public void setPublic(boolean isPublic) {
        this.isPublic = isPublic;
    }

    @Override
    public boolean isPublic() {
        return isPublic;
    }

    @Override
    public Collection<String> getPrefer() {
        return prefer;
    }

    @Override
    public void setDomain(Domain domain) {
        this.domain = domain;
    }

    @Override
    public Domain getDomain() {
        return domain;
    }

    @Override
    public String getDomainName() {
        return domain != null ? domain.getName() : ""; // use empty space for map key
    }

    @Override
    public void setSubLibrary(boolean isSubLibrary) {
        this.isSubLibrary = isSubLibrary;
    }

    @Override
    public boolean isSubLibrary() {
        return isSubLibrary;
    }

    @Override
    public void setPlace(Place place) {
        this.place = place;
        setPublic(place.isPublic());
    }

    @Override
    public Place getPlace() {
        return place;
    }

    @Override
    public void setGroups(Collection<String> groups) {
        this.groups = groups;
    }

    @Override
    public Collection<String> getGroups() {
        return groups;
    }

    @Override
    public Predicate<Library> getGroupPredicate() {
        return groupPredicate;
    }

    @Override
    public void setStations(Collection<Station> stations) {
        this.stations = stations;
    }

    @Override
    public Collection<Station> getStations() {
        return stations;
    }

    @Override
    public void setSelf(Collection<String> self) {
        this.self.clear();
        this.self.addAll(self);
    }

    @Override
    public Collection<String> getSelf() {
        return self;
    }

    @Override
    public void setSelfPredicate(Predicate<Library> selfPredicate) {
        this.selfPredicate = selfPredicate;
    }

    @Override
    public Predicate<Library> getSelfPredicate() {
        return selfPredicate;
    }

    @Override
    public void setLocal(Collection<String> local) {
        this.local.clear();
        this.local.addAll(local);
    }

    @Override
    public Collection<String> getLocal() {
        return local;
    }

    @Override
    public void setLocalPredicate(Predicate<Library> localPredicate) {
        this.localPredicate = localPredicate;
    }

    @Override
    public Predicate<Library> getLocalPredicate() {
        return localPredicate;
    }

    @Override
    public void setAccept(Collection<String> accept) {
        this.accept.clear();
        this.accept.addAll(accept);
    }

    @Override
    public Collection<String> getAccept() {
        return accept;
    }

    @Override
    public void setPrefer(Collection<String> prefer) {
        this.prefer.clear();
        this.prefer.addAll(prefer);
    }

    @Override
    public void setServicesPredicate(Predicate<LibraryService> servicesPredicate) {
        this.servicesPredicate = servicesPredicate;
    }

    @Override
    public Predicate<LibraryService> getServicesPredicate() {
        return servicesPredicate;
    }

    @Override
    public void setPriority(Integer priority) {
        this.priority = priority;
    }

    @Override
    public Integer getPriority() {
        return priority;
    }

    @Override
    public void setSenderAddress(LibrarySenderAddress senderAddress) {
        this.senderAddress = senderAddress;
    }

    @Override
    public LibrarySenderAddress getSenderAddress() {
        return senderAddress;
    }

    @Override
    public void setNotificationAddress(LibraryNotificationAddress notificationAddress) {
        this.notificationAddress = notificationAddress;
    }

    @Override
    public LibraryNotificationAddress getNotificationAddress() {
        return notificationAddress;
    }

    @Override
    public void setPostalAddress(LibraryPostalAddress postalAddress) {
        this.postalAddress = postalAddress;
    }

    @Override
    public LibraryPostalAddress getPostalAddress() {
        return postalAddress;
    }

    @Override
    public void setActiveEmail(String email) {
        this.activeEmail = email;
    }

    @Override
    public String getActiveEmail() {
        return activeEmail != null ? activeEmail :
                postalAddress != null ? postalAddress.getEmail() : null;
    }

    @Override
    public void setPassiveEmail(String email) {
        this.passiveEmail = email;
    }

    @Override
    public String getPassiveEmail() {
        return passiveEmail != null ? passiveEmail :
                postalAddress != null ? postalAddress.getEmail() : null;
    }

    @Override
    public void setReservationEmail(String email) {
        this.reservationEmail = email;
    }

    @Override
    public String getReservationEmail() {
        return reservationEmail != null ? reservationEmail :
                postalAddress != null ? postalAddress.getEmail() : null;
    }

    @Override
    public String toString() {
        return isil;
    }

    @Override
    public int compareTo(Library o) {
        return getISIL().compareTo(o.getISIL());
    }

    @Override
    public int hashCode() {
        return isil.hashCode();
    }

    @Override
    public boolean equals(Object object) {
        return object instanceof DefaultLibrary && isil.equals(((DefaultLibrary) object).getISIL());
    }

    @Override
    public Map<String, Object> asMap() {
        TinyMap.Builder<String, Object> map = TinyMap.builder();
        map.put("isil", getISIL());
        map.put("name", getName());
        map.put("fullname", getFullName());
        map.put("groups", getGroups());
        Collection<Station> stations = getStations();
        if (stations != null && !stations.isEmpty()) {
            TinyMultiMap<String, String> multiMap = new TinyMultiMap<>();
            for (String process : getProcesses(stations)) {
                multiMap.putAll(process, getNamesForProcess(stations, process));
            }
            map.put("locations", multiMap.asMap());
        }
        map.put("is_registered_interlibrary", isRegisteredInterlibrary());
        map.put("is_copy_interlibrary", isCopyInterlibrary());
        map.put("is_online_interlibrary", isOnlineInterlibrary());
        map.put("is_sublibrary", isSubLibrary());
        map.put("is_lastresort", isLastResort());
        map.put("is_organization", isOrganization());
        map.put("is_public", isPublic());
        Domain domain = getDomain();
        if (domain != null) {
            map.put("domain", domain.getName());
            map.put("organization", domain.getOrganization());
        }
        map.putIfNotNull("sigel", getSigel());
        map.putIfNotNull("ezb", getEzbId());
        map.put("aliases", getAliases());
        map.put("activeemail", getActiveEmail());
        map.put("passiveemail", getPassiveEmail());
        map.put("reservationemail", getReservationEmail());
        if (place != null) {
            map.put("country", place.getCountry());
            map.put("province", place.getProvince());
            map.put("place", place.getPlace());
            map.put("zip", place.getZip());
        }
        if (postalAddress != null) {
            map.put("postal", postalAddress.asMap());
        }
        if (senderAddress != null) {
            map.put("sender", senderAddress.asMap());
        }
        if (notificationAddress != null) {
            map.put("notification", notificationAddress.asMap());
        }
        map.put("self", getSelf());
        map.put("local", getLocal());
        map.put("accept", getAccept());
        map.put("prefer", getPrefer());
        return map.build();
    }

    @SuppressWarnings({"rawtypes", "unchecked"})
    protected void parseInternalMap(Map<String, Object> map) {
        if (map == null || map.isEmpty()) {
            return;
        }
        setISIL(Maps.getString(map, "isil"));
        setName(Maps.getString(map, "name"));
        setFullName(Maps.getString(map, "fullname"));
        setRegisteredInterlibrary(Maps.getBoolean(map, "is_registered_interlibrary"));
        setCopyInterlibrary(Maps.getBoolean(map, "is_copy_interlibrary"));
        setOnlineInterlibrary(Maps.getBoolean(map, "is_online_interlibrary"));
        setSubLibrary(Maps.getBoolean(map, "is_sublibrary"));
        setLastResort(Maps.getBoolean(map, "is_lastresort"));
        setOrganization(Maps.getBoolean(map, "is_organization"));
        setPublic(Maps.getBoolean(map, "is_public"));
        setGroups(Maps.getList(map, "groups").stream()
                .filter(Objects::nonNull)
                .map(Object::toString)
                .collect(Collectors.toList()));
        Map<?, ?> m = Maps.getMap(map, "locations");
        if (m != null && !m.isEmpty()) {
            List<Station> stations = new ArrayList<>();
            EnumSet.allOf(Station.Process.class).forEach(p -> {
                List<String> list = (List<String>) m.get(p.name().toLowerCase(Locale.ROOT));
                if (list != null && !list.isEmpty()) {
                    for (String string : list) {
                        Station station = DefaultStation.builder()
                                .setProcess(p)
                                .setISIL(isil)
                                .setName(string)
                                .build();
                        stations.add(station);
                    }
                }
            });
            if (!stations.isEmpty()) {
                setStations(stations);
            }
        }
        setDomain(Domain.of(Maps.getString(map,"domain"), Maps.getString(map, "organization")));
        setSigel(Maps.getString(map, "sigel"));
        setEzbId(Maps.getString(map, "ezb"));
        setAliases(Maps.getList(map, "aliases").stream()
                .filter(Objects::nonNull)
                .map(Object::toString)
                .collect(Collectors.toList()));
        setActiveEmail(Maps.getString(map, "activeemail"));
        setPassiveEmail(Maps.getString(map, "passiveemail"));
        setReservationEmail(Maps.getString(map, "reservationemail"));
        String country = Maps.getString(map, "country");
        String province = Maps.getString(map, "province");
        String place = Maps.getString(map, "place");
        String zip = Maps.getString(map, "zip");
        setPlace(Place.of(country, province, place, zip, isPublic()));
        DefaultLibraryPostalAddress postalAddress = new DefaultLibraryPostalAddress();
        postalAddress.setName(Maps.getString(map, "postal.name"));
        postalAddress.setShortName(Maps.getString(map, "postal.shortname"));
        postalAddress.setStreet(Maps.getString(map, "postal.street"));
        postalAddress.setZipCode(Maps.getString(map, "postal.zip"));
        postalAddress.setCountry(Maps.getString(map, "postal.country"));
        postalAddress.setProvince(Maps.getString(map, "postal.province"));
        postalAddress.setPostOfficeBox(Maps.getString(map, "postal.postofficebox"));
        postalAddress.setPostOfficeBoxZipCode(Maps.getString(map, "postal.postofficeboxzip"));
        postalAddress.setPlace(Maps.getString(map, "postal.place"));
        boolean postalpublic = Maps.getBoolean(map, "postal.public");
        postalAddress.setPublic(postalpublic);
        postalAddress.setEmail(Maps.getString(map, "postal.email"));
        Map emails = Maps.getMap(map, "postal.emails").entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue,
                        (u, v) -> { throw new IllegalStateException(); }, TinyMap::builder));
        postalAddress.setEmails(emails);
        Map phoneMap = Maps.getMap(map, "postal.phone").entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue,
                        (u, v) -> { throw new IllegalStateException(); }, TinyMap::builder));
        postalAddress.setPhoneNumber(phoneMap);
        Maps.getList(map, "postal.container").stream()
                .filter(Objects::nonNull)
                .map(Object::toString)
                .toList().forEach(postalAddress::addContainerInformation);
        postalAddress.setHomePageURL(Maps.getString(map, "postal.homepage"));
        postalAddress.setOpeningHours(Maps.getString(map, "postal.openinghours"));
        postalAddress.setOpeningHoursComments(Maps.getString(map, "postal.openinghourscomments"));
        postalAddress.setLatitude(Maps.getString(map, "postal.latitude"));
        postalAddress.setLongitude(Maps.getString(map, "postal.longitude"));
        setPostalAddress(postalAddress);
        DefaultLibrarySenderAddress senderAddress = new DefaultLibrarySenderAddress();
        senderAddress.setName(Maps.getString(map, "sender.name"));
        senderAddress.setStreet(Maps.getString(map, "sender.street"));
        senderAddress.setZipCode(Maps.getString(map, "sender.zip"));
        senderAddress.setPlace(Maps.getString(map, "sender.place"));
        senderAddress.setCountry(Maps.getString(map, "sender.country"));
        senderAddress.setProvince(Maps.getString(map, "sender.province"));
        senderAddress.setPostOfficeBox(Maps.getString(map, "sender.postofficebox"));
        senderAddress.setPostOfficeBoxZipCode(Maps.getString(map, "sender.postofficeboxzip"));
        setSenderAddress(senderAddress);
        DefaultLibraryNotificationAddress notificationAddress = new DefaultLibraryNotificationAddress();
        notificationAddress.setName(Maps.getString(map, "notification.name"));
        notificationAddress.setShortName(Maps.getString(map, "notification.shortname"));
        notificationAddress.setStreet(Maps.getString(map, "notification.street"));
        notificationAddress.setZipCode(Maps.getString(map, "notification.zip"));
        notificationAddress.setPlace(Maps.getString(map, "notification.place"));
        notificationAddress.setCountry(Maps.getString(map, "notification.country"));
        notificationAddress.setProvince(Maps.getString(map, "notification.province"));
        notificationAddress.setEmail(Maps.getString(map, "notification.email"));
        Map notificationemails = Maps.getMap(map, "notification.emails").entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue,
                        (u, v) -> { throw new IllegalStateException(); }, TinyMap::builder));
        notificationAddress.setEmails(notificationemails);
        phoneMap = Maps.getMap(map, "notification.phone").entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue,
                        (u, v) -> { throw new IllegalStateException(); }, TinyMap::builder));
        notificationAddress.setPhoneNumber(phoneMap);
        notificationAddress.setHomePageURL(Maps.getString(map, "notification.homepage"));
        notificationAddress.setOpeningHours(Maps.getString(map, "notification.openinghours"));
        notificationAddress.setOpeningHoursComments(Maps.getString(map, "notification.openinghourscomments"));
        notificationAddress.setLatitude(Maps.getString(map, "notification.latitude"));
        notificationAddress.setLongitude(Maps.getString(map, "notification.longitude"));
        setNotificationAddress(notificationAddress);
        Collection<String> self = Maps.getList(map, "self");
        setSelf(self);
        setSelfPredicate(p -> self.contains(p.getISIL()));
        Collection<String> local = Maps.getList(map, "local");
        setLocal(local);
        setLocalPredicate(p -> local.contains(p.getISIL()));
        Collection<String> accept = Maps.getList(map, "accept");
        setAccept(accept);
        Collection<String> prefer = Maps.getList(map, "prefer");
        setPrefer(prefer);
    }

    private static Collection<String> getProcesses(Collection<Station> stations) {
        return stations.stream().map(station -> station.getProcess().name().toLowerCase(Locale.ROOT))
                .sorted().distinct()
                .toList();
    }

    private static Collection<String> getNamesForProcess(Collection<Station> stations, String process) {
        return stations.stream().filter(station -> process.equals(station.getProcess().name().toLowerCase(Locale.ROOT)))
                .map(Station::getName)
                .toList();
    }
}
