package org.xbib.interlibrary.api.library;

import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.xbib.datastructures.tiny.TinyList;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.api.util.Maps;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.TEXT;

public class DefaultLibraryService implements LibraryService, Comparable<LibraryService> {

    private static final Random random = new SecureRandom();

    private final Builder builder;

    public DefaultLibraryService(Builder builder) {
        this.builder = builder;
    }

    public static Builder builder() {
        return new Builder();
    }

    public void setISIL(String isil) {
        builder.isil = isil;
    }

    @Override
    public String getId() {
        return builder.id;
    }

    @Override
    public Integer getPriority() {
        return builder.priority;
    }

    @Override
    public Map<String, Object> asMap() {
        TinyMap.Builder<String, Object> map = TinyMap.builder();
        if (builder.id != null) {
            map.put("id", builder.id);
        }
        if (builder.isil != null) {
            map.put("isil", builder.isil);
        }
        if (builder.name != null) {
            map.put("name", builder.name);
        }
        if (builder.priority != null) {
            map.put("priority", builder.priority);
        }
        map.put("material", builder.materials.stream()
                .map(Material::toValue)
                .toList());
        if (builder.mode != null) {
            map.put("mode", builder.mode.toValue());
        }
        if (!builder.locations.isEmpty()) {
            TinyList.Builder<Map<String, Object>> list = TinyList.builder();
            for (Location location : builder.locations) {
                TinyMap.Builder<String, Object> m = TinyMap.builder();
                if (location.getLibrary() != null) {
                    m.put("library", location.getLibrary());
                }
                if (location.getLibraryName() != null) {
                    m.put("libraryname", location.getLibraryName());
                }
                if (location.getCollection() != null) {
                    m.put("collection", location.getCollection());
                }
                if (location.getCallnumber() != null) {
                    m.put("callnumber", location.getCallnumber());
                }
                if (location.getDescription() != null) {
                    m.put("description", location.getDescription());
                }
                if (location.getTextualHoldings() != null) {
                    m.put("textualholdings", location.getTextualHoldings());
                }
                if (!location.getLinks().isEmpty()) {
                    TinyList.Builder<Map<String, Object>> l = TinyList.builder();
                    location.getLinks().forEach(lu -> {
                        TinyMap.Builder<String, Object> link = TinyMap.builder();
                        if (lu.getUrl() != null) {
                            link.put("url", lu.getUrl());
                        }
                        if (lu.getLabel() != null) {
                            link.put("label", lu.getLabel());
                        }
                        l.add(link);
                    });
                    m.put("links", l);
                }
                list.add(m.build());
            }
            map.put("location", list.build());
        }
        if (builder.source != null) {
            map.put("source", builder.source);
        }
        if (builder.sourceId != null) {
            map.put("sourceId", builder.sourceId);
        }
        return map.build();
    }

    @Override
    public String getISIL() {
        return builder.isil;
    }

    @Override
    public String getName() {
        return builder.name;
    }

    @Override
    public String getKey() {
        return builder.key;
    }

    @Override
    public List<Material> getMaterials() {
        return builder.materials;
    }

    @Override
    public Mode getMode() {
        return builder.mode;
    }

    @Override
    public List<Location> getLocations() {
        return builder.locations;
    }

    @Override
    public boolean containsYear(Integer year) {
        return builder.years.contains(year);
    }

    @Override
    public boolean isValid() {
        return builder.isil != null;
    }

    @Override
    public CharSequence getRandomizedKey() {
        return builder.randomKey;
    }

    @Override
    public CharSequence getSimpleKey() {
        return builder.simpleKey;
    }

    @Override
    public int hashCode() {
        return builder.toString().hashCode();
    }

    @Override
    public boolean equals(Object object) {
        return object instanceof LibraryService && toString().equals(object.toString());
    }

    @Override
    public int compareTo(LibraryService o) {
        return toString().compareTo(o.toString());
    }

    @Override
    public String toString() {
        // for debug only
        return getISIL() + "|" + getId() + "|" + getMode() + "|" + getMaterials() + "|" +
                getLocations() +
                "[prio=" + getPriority() + "]" +
                "[key=" + getSimpleKey() + "]";
    }

    public void appendLocation(List<Location> locations) {
        if (locations != null) {
            for (Location location : locations) {
                appendLocation(location);
            }
        }
    }

    public void appendLocation(Location locationToAppend) {
        if (getLocations().isEmpty()) {
            getLocations().add(locationToAppend);
            return;
        }
        DefaultLocation lastLocation = (DefaultLocation) (getLocations().getLast());
        lastLocation.setCallnumber(append(lastLocation.getCallnumber(), locationToAppend.getCallnumber()));
        lastLocation.setDescription(appendWithDuplicate(lastLocation.getDescription(), locationToAppend.getDescription()));
        lastLocation.setTextualHoldings(appendWithDuplicate(lastLocation.getTextualHoldings(), locationToAppend.getTextualHoldings()));
    }

    private static String append(String string1, String string2) {
        if (string1 != null && string2 != null) {
            if (!string1.contains(string2)) {
                return string1 + ", " + string2;
            } else {
                return string1;
            }
        }
        if (string1 != null) {
            return string1;
        }
        return string2;
    }

    private static String appendWithDuplicate(String string1, String string2) {
        if (string1 != null && string2 != null) {
            return string1 + ", " + string2;
        }
        if (string1 != null) {
            return string1;
        }
        return string2;
    }

    public static class Builder implements LibraryService.Builder {

        private String id;

        private String isil;

        private String name;

        private String key;

        private String source;

        private String sourceId;

        private Integer domainPriority;

        private Integer libraryPriority;

        private Integer priority;

        private Integer subPriority;

        private StringBuilder simpleKey;

        private CharSequence randomKey;

        private Mode mode;

        private List<Material> materials;

        private final LinkedList<Location> locations;

        private final List<Integer> years;

        private Builder() {
            this.materials = new ArrayList<>();
            this.locations = new LinkedList<>();
            this.years = new ArrayList<>();
            this.domainPriority = INVALID_PRIORITY;
            this.libraryPriority = INVALID_PRIORITY;
            this.priority = INVALID_PRIORITY;
            this.subPriority = INVALID_PRIORITY;
            this.mode = Mode.UNSPECIFIED;
        }

        @SuppressWarnings("unchecked")
        public Builder parse(Map<String, Object> map) {
            if (map.containsKey("id")) {
                setId((String) map.get("id"));
            }
            if (map.containsKey("isil")) {
                setISIL((String) map.get("isil"));
            }
            if (map.containsKey("name")) {
                setName((String) map.get("name"));
            }
            Long priority = (Long) map.get("priority");
            if (priority != null) {
                setServicePriority(priority.intValue());
            }
            if (map.containsKey("mode")) {
                Object object = map.get("mode");
                if (object instanceof Collection<?> collection) {
                    setMode(collection.isEmpty() ? Mode.UNDEFINED :
                                Mode.fromValue(collection.iterator().next().toString()));
                } else if (object instanceof String) {
                    setMode(Mode.fromValue((String) object));
                }
            }
            if (map.containsKey("material")) {
                Object object = map.get("material");
                if (object instanceof Collection<?> collection) {
                    setMaterials(collection.stream().map(Object::toString).map(Material::fromValue).toList());
                } else if (object instanceof String) {
                    setMaterials(List.of(Material.fromValue((String) object)));
                }
            }
            if (map.containsKey("source")) {
                setSource((String) map.get("source"));
            }
            if (map.containsKey("sourceid")) {
                setSourceId((String) map.get("sourceid"));
            }
            setLocations(createLocations((List<Map<String, Object>>) map.get("location")));
            return this;
        }

        private List<Location> createLocations(List<Map<String, Object>> maps) {
            if (maps == null) {
                return List.of();
            }
            List<Location> list = new ArrayList<>();
            for (Map<String, Object> map : maps) {
                DefaultLocation location = (DefaultLocation) DefaultLocation.parse(map);
                list.add(location);
            }
            return list;
        }

        @Override
        public Builder setId(String id) {
            this.id = id;
            return this;
        }

        @Override
        public Builder setISIL(String isil) {
            this.isil = isil;
            return this;
        }

        @Override
        public Builder setName(String name) {
            this.name = name;
            return this;
        }

        @Override
        public Builder setKey(String key) {
            this.key = key;
            return this;
        }

        public Builder setSource(String source) {
            this.source = source;
            return this;
        }

        public Builder setSourceId(String sourceId) {
            this.sourceId = sourceId;
            return this;
        }

        @Override
        public Builder setDomainPriority(Integer domainPriority) {
            this.domainPriority = domainPriority;
            return this;
        }

        @Override
        public Builder setLibraryPriority(Integer libraryPriority) {
            this.libraryPriority = libraryPriority;
            return this;
        }

        @Override
        public Builder setServicePriority(Integer priority) {
            this.priority = priority;
            return this;
        }

        @Override
        public Builder setSublibraryPriority(Integer subPriority) {
            this.subPriority = subPriority;
            return this;
        }

        @Override
        public Builder setMode(Mode mode) {
            this.mode = mode;
            return this;
        }

        @Override
        public Builder addLocation(Location location) {
            this.locations.add(location);
            return this;
        }

        @Override
        public Builder setMaterials(List<Material> materials) {
            this.materials = materials;
            return this;
        }

        public Builder setLocations(List<Location> locations) {
            if (locations != null) {
                this.locations.addAll(locations);
            }
            return this;
        }

        public Builder setYears(List<Integer> years) {
            if (years != null) {
                this.years.addAll(years);
            }
            return this;
        }

        @Override
        public LibraryService build() {
            this.simpleKey = new StringBuilder()
                    .append(String.format("%02d", domainPriority))
                    .append(String.format("%02d", libraryPriority))
                    .append(String.format("%02d", priority))
                    .append(String.format("%02d", subPriority));
            this.randomKey = new StringBuilder(simpleKey)
                    .append(String.format("%04d", random.nextInt(10000)));
            return new DefaultLibraryService(this);
        }
    }

    public static class DefaultLocation implements Location {

        private String library;

        private String libraryname;

        private String callnumber;

        private String collection;

        private String description;

        private String textualHoldings;

        private final List<LabeledUrl> links;

        public DefaultLocation() {
            this.links = new ArrayList<>();
        }

        @SuppressWarnings("unchecked")
        public static Location parse(Map<String, Object> map) {
            DefaultLocation location = new DefaultLocation();
            // : class java.lang.StringBuilder cannot be cast to class java.lang.String (java.lang.StringBuilder and java.lang.String are in module java.base of loader 'bootstrap')
            location.setLibrary(Maps.getString(map,"library"));
            location.setLibraryName(Maps.getString(map,"libraryname"));
            String collection = Maps.getString(map,"collection");
            location.setCollection(clean(collection));
            String callnumber = Maps.getString(map,"callnumber");
            location.setCallnumber(clean(callnumber));
            String description = Maps.getString(map, "description");
            location.setDescription(clean(description));
            String textualHoldings = Maps.getString(map,"textualholdings");
            location.setTextualHoldings(clean(textualHoldings));
            List<Map<String, Object>> links = (List<Map<String, Object>>) map.get("link");
            if (links != null) {
                for (Map<String, Object> link : links) {
                    location.addLink((String) link.get("url"), (String) link.get("label"));
                }
            }
            return location;
        }

        public void setLibrary(String library) {
            this.library = library;
        }

        public void setLibraryName(String libraryName) {
            this.libraryname = libraryName;
        }

        public void setCollection(String collection) {
            this.collection = collection;
        }

        public void setCallnumber(String callnumber) {
            this.callnumber = callnumber;
        }

        public void setTextualHoldings(String textualHoldings) {
            this.textualHoldings = textualHoldings;
        }

        public void setDescription(String description) {
            this.description = description;
        }

        public void addLink(LabeledUrl labeledUrl) {
            this.links.add(labeledUrl);
        }

        public void addLink(String url, String label) {
            this.links.add(new DefaultLabeledUrl().setUrl(url).setLabel(label));
        }

        @Override
        public String getLibrary() {
            return library;
        }

        @Override
        public String getLibraryName() {
            return libraryname;
        }

        @Override
        public String getCollection() {
            return collection;
        }

        @Override
        public String getCallnumber() {
            return callnumber;
        }

        @Override
        public String getDescription() {
            return description;
        }

        @Override
        public String getTextualHoldings() {
            return textualHoldings;
        }

        @Override
        public List<LabeledUrl> getLinks() {
            return links;
        }

        public String toString() {
            return collection + "/" + callnumber + "/" + description + "/" + textualHoldings;
        }
    }

    public static class DefaultLabeledUrl implements LabeledUrl {

        private String url;

        private String label;

        public DefaultLabeledUrl() {
        }

        public DefaultLabeledUrl setUrl(String url) {
            this.url = url;
            return this;
        }

        public DefaultLabeledUrl setLabel(String label) {
            this.label = label;
            return this;
        }

        @Override
        public String getUrl() {
            return url;
        }

        @Override
        public String getLabel() {
            return label;
        }
    }

    public static final LibraryService INTERLIBRARY_COPY_ONLINE = DefaultLibraryService.builder()
            .setMode(Mode.COPY)
            .setMaterials(List.of(ONLINE_RESOURCE))
            .build();

    public static final LibraryService INTERLIBRARY_COPY_PRINT = DefaultLibraryService.builder()
            .setMode(Mode.COPY)
            .setMaterials(List.of(TEXT))
            .build();

    private static String clean(String string) {
        if (string == null) {
            return null;
        }
        return string.replaceAll("\u0098", "")
                .replaceAll("\u009c", "")
                .replaceAll("\\p{C}", "")
                .trim();
    }
}
