package org.xbib.interlibrary.api.library;

import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import org.xbib.datastructures.tiny.TinyMap;

public class DefaultStation implements Station {

    private final Builder builder;

    private DefaultStation(Builder builder) {
        this.builder = builder;
    }

    public static Builder builder() {
        return new Builder();
    }

    @Override
    public boolean isEnabled() {
        return builder.enabled;
    }

    @Override
    public String getISIL() {
        return builder.isil;
    }

    @Override
    public String getName() {
        return builder.name;
    }

    @Override
    public Process getProcess() {
        return builder.process;
    }

    @Override
    public String getSortKey() {
        return builder.sortkey;
    }

    @Override
    public List<URI> getURIs() {
        return builder.uris;
    }

    @Override
    public String toString() {
        return "[Station:" +
                "enabled=" + isEnabled() +
                ",process=" + getProcess() +
                ",isil=" + getISIL() +
                ",name=" + getName() +
                ",sortkey=" + getSortKey() +
                ",uris=" + getURIs() + "]";
    }

    @Override
    public Map<String, Object> asMap() {
        TinyMap.Builder<String, Object> map = TinyMap.builder();
        map.put("enabled", builder.enabled);
        map.put("process", builder.process.name());
        map.put("isil", builder.isil);
        map.put("name", builder.name);
        map.put("sortkey", builder.sortkey);
        map.put("uris", builder.uris.stream().map(URI::toString).collect(Collectors.joining(",")));
        return map.build();
    }

    @SuppressWarnings("unchecked")
    public static Station fromMap(Map<String, Object> map) {
        Builder builder = builder();
        if (map.containsKey("enabled")) {
            builder.setEnabled((boolean) map.get("enabled"));
        }
        if (map.containsKey("process")) {
            builder.setProcess(Process.valueOf((String) map.get("process")));
        }
        if (map.containsKey("isil")) {
            builder.setISIL((String) map.get("isil"));
        }
        if (map.containsKey("name")) {
            builder.setName((String) map.get("name"));
        }
        if (map.containsKey("sortkey")) {
            builder.setSortKey((String) map.get("sortkey"));
        }
        if (map.containsKey("uris")) {
            String uriString = (String) map.get("uris");
            builder.setURIs(Arrays.stream(uriString.split(",")).map(URI::create).toList());
        }
        return builder.build();
    }

    public static class Builder implements Station.Builder {

        private boolean enabled;

        private Station.Process process;

        private String isil;

        private String name;

        private String sortkey;

        private List<URI> uris;

        private Builder() {
            this.uris = new ArrayList<>();
        }

        public Builder setEnabled(boolean enabled) {
            this.enabled = enabled;
            return this;
        }

        public Builder setProcess(Process process) {
            this.process = process;
            return this;
        }

        public Builder setISIL(String isil) {
            this.isil = isil;
            return this;
        }

        public Builder setName(String name) {
            this.name = name;
            return this;
        }

        public Builder setSortKey(String sortkey) {
            this.sortkey = sortkey;
            return this;
        }

        public Builder setURIs(List<URI> uris) {
            if (uris != null) {
                this.uris = uris;
            }
            return this;
        }

        @Override
        public Station build() {
            return new DefaultStation(this);
        }
    }
}
